/**
 * --------------------------------------------------------------------------------
 *  NoiseTube Mobile client (Java implementation)
 *  
 *  Copyright (C) 2008-2010 SONY Computer Science Laboratory Paris
 *  Portions contributed by Vrije Universiteit Brussel (BrusSense team), 2008-2011
 * --------------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU Lesser General Public License, version 2.1, as published
 *  by the Free Software Foundation.
 *  
 *  This library is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License along
 *  with this library; if not, write to:
 *    Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor,
 *    Boston, MA  02110-1301, USA.
 *  
 *  Full GNU LGPL v2.1 text: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
 *  NoiseTube project source code repository: http://code.google.com/p/noisetube
 * --------------------------------------------------------------------------------
 *  More information:
 *   - NoiseTube project website: http://www.noisetube.net
 *   - Sony Computer Science Laboratory Paris: http://csl.sony.fr
 *   - VUB BrusSense team: http://www.brussense.be
 * --------------------------------------------------------------------------------
 */

package net.noisetube.core;

import java.util.Vector;

import net.noisetube.audio.SoundLevelMeter;
import net.noisetube.audio.calibration.Calibration;
import net.noisetube.config.Preferences;
import net.noisetube.io.saving.FileSaver;
import net.noisetube.io.saving.HttpSaver;
import net.noisetube.io.saving.MultiSaver;
import net.noisetube.io.saving.Saver;
import net.noisetube.location.GeoTagger;
import net.noisetube.model.SLMMeasurement;
import net.noisetube.model.NTMeasurement;
import net.noisetube.model.NTCoordinates;
import net.noisetube.model.MeasurementListener;
import net.noisetube.model.Track;
import net.noisetube.util.Logger;
import net.noisetube.util.StringUtils;

/**
 * @author mstevens
 *
 */
public abstract class NTClient extends SLMClient
{

	//STATIC-----------------------------------------------
	static final public int EMULATOR_ENV = 0; 		//Emulator environment
	static final public int PHONE_DEV_ENV = 1; 		//Development phone environment
	static final public int PHONE_PROD_ENV = 2; 	//Production phone environment
	
	//Set active environment:
	static public int ENVIRONMENT = PHONE_PROD_ENV; //default
	
	public static String getEnvironmentName(int environment)
	{
		String name = null;
    	switch(environment)
    	{
    		case PHONE_PROD_ENV : name = "Phone/Production"; break;
    		case PHONE_DEV_ENV : name = "Phone/Development"; break;
    		case EMULATOR_ENV : name = "Emulator"; break;
	        default : name = "unknown";
    	}
    	return name;
	}
	
	public static NTClient getInstance()
	{
		return (NTClient) INSTANCE; //!!!
	}
	
	public static void dispose()
	{
		Logger.dispose();
		INSTANCE = null;
	}
	
	//DYNAMIC----------------------------------------------
	protected Preferences preferences;
	protected GeoTagger geoTagger;
	protected Saver saver;
	
	protected NTClient(String clientType, String clientVersion, String clientBuildDate, boolean testVersion)
	{
		super(clientType, clientVersion, clientBuildDate, testVersion);
		
		//Subclasses must call initialize!
	}
	
	/**
	 * Sets the Device and Preferences
	 */
	protected final void initialize() throws Exception
	{
		if(!initialized)
		{
			super.initialize(); //!!!
			
			//Check baseline functionality: reading/writing files & recording audio
			Vector missingFeatures = new Vector();
			if(!device.supportsFileAccess())
				missingFeatures.addElement("Cannot access filesystem");
			if(!device.supportsAudioRecording())
				missingFeatures.addElement("Cannot record audio");
			
			if(!missingFeatures.isEmpty())
			{
				String errorMsg = "Unsupported device! Required functionality missing: " + StringUtils.StringVectorToString(missingFeatures, "; ");
				//write crash log to file if we can
				try
				{
					if(device.supportsFileAccess())
					{
						log.error(errorMsg);
						log.dumpCrashLog(createPreferences().getDataFolderPath());
					}
				}
				catch(Exception ignore) { } 
				throw new Exception(errorMsg); //!!! (will show up in GUI)
			}
			
			//Create preferences:
			preferences = createPreferences();
			
			//Last but not least:
			initialized = true;
		}
	}
	
	/**
	 * @return the preferences
	 */
	public Preferences getPreferences()
	{
		return preferences;
	}
	
	protected abstract Preferences createPreferences();

	/**
	 * @param audioSpec
	 * @param listener
	 * @return an slm instance with a calibration (but not necessarily a correct/fitting one) 
	 * @throws Exception
	 * 
	 * the difference with SLMClient.getCalibratedSLM() is that here we use the calibration from pref's, not from device
	 */
	public SoundLevelMeter getCalibratedSLM(MeasurementListener listener) throws Exception
	{
		if(slm == null)
			slm = new SoundLevelMeter(device.getAudioSpecification(), preferences.getCalibration(), listener);
		else
			slm.setListener(listener);
		return slm;
	}
	
	public GeoTagger getGeoTagger()
	{
		if(geoTagger == null)
			geoTagger = createGeoTagger();
		return geoTagger;
	}
	
	protected abstract GeoTagger createGeoTagger();
	
	public Saver getSaver(Track track)
	{
		if(saver == null)
		{
			//Create a saver...
			switch(preferences.getSavingMode())
			{
				case Preferences.SAVE_NO : saver = null; break;
				case Preferences.SAVE_HTTP:
					if(preferences.isAlsoSaveToFileWhenInHTTPMode())
					{
						saver = new MultiSaver(track);
						((MultiSaver) saver).addSaver(new HttpSaver(track));
						((MultiSaver) saver).addSaver(new FileSaver(track));
					}
					else
						saver = new HttpSaver(track);
					break;
				case Preferences.SAVE_FILE:
					saver = new FileSaver(track);
					break;
				case Preferences.SAVE_SMS: /* ... */
					break;
					//...
			}
		}
		else
			saver.setTrack(track); //!!! assign saver to given track
		return saver;
	}
	
	/**
	 * To be called from Preferences upon a change of a saving preference
	 */
	public void resetSaver()
	{
		saver = null;
	}
	
	public abstract NTCoordinates getNTCoordinates(double latitude, double longitude, double altitude);
	
	public abstract void addTrackProcessors(Track track);
	
	public void annotateTrack(Track track)
	{
		track.addMetadata("client", clientType);
		track.addMetadata("clientVersion", clientVersion);
		track.addMetadata("clientBuildDate", clientBuildDate);
		track.addMetadata("deviceBrand", device.getBrand());
		track.addMetadata("deviceModel", device.getModel());
		track.addMetadata("deviceModelVersion", device.getModelVersion());
		track.addMetadata("devicePlatform", device.getPlatform());
		track.addMetadata("devicePlatformVersion", device.getPlatformVersion());
		track.addMetadata("deviceJavaPlatform", device.getJavaPlatform());
		track.addMetadata("deviceJavaPlatformVersion", device.getJavaPlatformVersion());
		if(preferences.getCalibration() != null)
		{
			track.addMetadata("calibration", preferences.getCalibration().toString());
			track.addMetadata("credibility", new String(new char[] {  preferences.getCalibration().getEffeciveCredibilityIndex() }));
		}
		else
			track.addMetadata("credibility", new String(new char[] { Calibration.CREDIBILITY_INDEX_X }));
		//subclasses can add some more:
		additionalTrackAnnotating(track);
	}
	
	public SLMMeasurement createMeasurement(long timeStampMS)
	{
		return new NTMeasurement(timeStampMS);
	}

	/**
	 * For audio classes only (they do not know about preferences)
	 */
	public String getDataFolderPath()
	{
		return preferences.getDataFolderPath();
	}
	
	protected abstract void additionalTrackAnnotating(Track track);
	
}
