/**
 * --------------------------------------------------------------------------------
 *  NoiseTube Mobile client (Java implementation; Android version)
 *  
 *  Copyright (C) 2008-2010 SONY Computer Science Laboratory Paris
 *  Portions contributed by Vrije Universiteit Brussel (BrusSense team), 2008-2011
 *  Android port by Vrije Universiteit Brussel (BrusSense team), 2010-2011
 * --------------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU Lesser General Public License, version 2.1, as published
 *  by the Free Software Foundation.
 *  
 *  This library is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License along
 *  with this library; if not, write to:
 *    Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor,
 *    Boston, MA  02110-1301, USA.
 *  
 *  Full GNU LGPL v2.1 text: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
 *  NoiseTube project source code repository: http://code.google.com/p/noisetube
 * --------------------------------------------------------------------------------
 *  More information:
 *   - NoiseTube project website: http://www.noisetube.net
 *   - Sony Computer Science Laboratory Paris: http://csl.sony.fr
 *   - VUB BrusSense team: http://www.brussense.be
 * --------------------------------------------------------------------------------
 */

package net.noisetube.io;

import java.io.InputStream;

import net.noisetube.audio.calibration.CalibrationFactory.CalibrationsList;
import net.noisetube.audio.calibration.Calibration;
import net.noisetube.audio.calibration.CalibrationsParser;
import net.noisetube.core.SLMClient;
import net.noisetube.util.Logger;

/**
 * Only for SLM library, only includes pinging and calibration downloading
 * 
 * @author mstevens
 *
 */
public class SLMWebAPI
{

	private static int PING_TIMEOUT_SECONDS = 15;
	protected static final String AGENT = SLMClient.getInstance().getClientType() + "/" + SLMClient.getInstance().getClientVersion();
	static final String DEFAULT_API_BASE_URL = "http://www.noisetube.net/api/";
	static final String DEV_API_BASE_URL = "http://localhost:3000/api/";
	protected Logger log = Logger.getInstance();
	protected String apiBaseURL = DEFAULT_API_BASE_URL;
	protected HttpClient httpClient;

	public SLMWebAPI()
	{
		httpClient = SLMClient.getInstance().getHttpClient(AGENT);
	}

	public synchronized boolean ping()
	{
		try
		{
			Pinger pinger = new Pinger();
			Thread pingThread = new Thread(pinger);
			pingThread.start(); //async
			this.wait(PING_TIMEOUT_SECONDS * 1000);
			try
			{
				pingThread.interrupt();
			}
			catch(Exception ignore) {}
			if(pinger.response == null)
				log.debug("Ping failed or timed out");
			return(pinger.response != null && pinger.response.equalsIgnoreCase("ok"));
		}
		catch(Exception e)
		{
			log.error(e, "Upon pinging (async)"); //should never happen (exceptions are caught in Pinger) but kept the try/catch just to be sure
			return false; //ping failed
		}
	}

	public CalibrationsList downloadCalibrations(final CalibrationsParser parser)
	{
		try
		{
			DownloadedCalibrationsReader reader = new DownloadedCalibrationsReader(parser);
			httpClient.getRequest(apiBaseURL + "mobilecalibrations", reader);
			return reader.calibrationsList;
		}
		catch(Exception e)
		{
			log.error(e, "Could not download calibrations XML file from server");
			return null;
		}
	}
	
	/**
	 * @author mstevens
	 */
	class Pinger implements Runnable
	{
		
		String response = null;

		public void run()
		{
			try
			{
				response = httpClient.getRequest(apiBaseURL + "ping");
			}
			catch(Exception ignore) {}
			finally
			{
				synchronized(SLMWebAPI.this)
				{
					SLMWebAPI.this.notify(); //!!!
				}
				
			}
		}
		
	}
	
	/**
	 * @author mstevens
	 */
	class DownloadedCalibrationsReader implements IInputStreamReader
	{
		
		private CalibrationsParser parser;
		CalibrationsList calibrationsList = null;
		
		public DownloadedCalibrationsReader(CalibrationsParser parser)
		{
			this.parser = parser;
		}
		
		public void read(InputStream inputStream) throws Exception
		{
			calibrationsList = parser.parseList(inputStream, Calibration.SOURCE_DOWNLOADED); //closes the stream
		}
		
	}


}