/**
 * --------------------------------------------------------------------------------
 *  NoiseTube Mobile client (Java implementation)
 *  
 *  Copyright (C) 2008-2010 SONY Computer Science Laboratory Paris
 *  Portions contributed by Vrije Universiteit Brussel (BrusSense team), 2008-2011
 * --------------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU Lesser General Public License, version 2.1, as published
 *  by the Free Software Foundation.
 *  
 *  This library is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License along
 *  with this library; if not, write to:
 *    Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor,
 *    Boston, MA  02110-1301, USA.
 *  
 *  Full GNU LGPL v2.1 text: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
 *  NoiseTube project source code repository: http://code.google.com/p/noisetube
 * --------------------------------------------------------------------------------
 *  More information:
 *   - NoiseTube project website: http://www.noisetube.net
 *   - Sony Computer Science Laboratory Paris: http://csl.sony.fr
 *   - VUB BrusSense team: http://www.brussense.be
 * --------------------------------------------------------------------------------
 */

package net.noisetube.model;

import java.util.Vector;

import net.noisetube.tagging.TagsHelper;
import net.noisetube.util.CustomStringBuffer;
import net.noisetube.util.URLUTF8Encoder;
import net.noisetube.util.XMLUtils;

/**
 * NoiseTube Measurement Model
 * 
 * @author maisonneuve, mstevens, sbarthol
 *
 */
public class NTMeasurement extends SLMMeasurement implements Saveable
{
	
	private Vector userTags;
	private Vector automaticTags;
	NTLocation location = null;
 
	public NTMeasurement()
	{
		super();
	}
	
	public NTMeasurement(long timeStampMS)
	{
		super(timeStampMS);
	}
	
	public boolean hasTags()
	{
		return (userTags != null || automaticTags != null);
	}
	
	public boolean hasUserTags()
	{
		return (userTags != null);
	}
	
	public boolean hasAutomaticTags()
	{
		return (automaticTags != null);
	}
	
	public void addAutomaticTag(String tag)
	{
		if(automaticTags == null)
			automaticTags = new Vector();
		automaticTags.addElement(tag);
	}
	
	public void addUserTags(String userTypedTags)
	{
		String[] tags = TagsHelper.splitTypedTags(userTypedTags);
		if(tags.length > 0)
		{
			if(userTags == null)
				userTags = new Vector();
			for(int t = 0; t < tags.length; t++)
				userTags.addElement(tags[t].trim());
		}
	}
	
	/**
	 * @return the location
	 */
	public NTLocation getLocation()
	{
		return location;
	}
	/**
	 * @param location the location to set
	 */
	public void setLocation(NTLocation location)
	{
		this.location = location;
	}
	
	/**
	 * String representation
	 */
	public String toString()
	{
		return "date: " + timeStamp + (location != null ? "; location: " + location.toString() : "") + "; Leq: " + LeqDBA + " dB(A)";
	}
	
	/**
	 * URL representation
	 */
	public String toUrl()
	{
		StringBuffer bff = new StringBuffer();
		bff.append("db=" + (int)LeqDBA.doubleValue());
		bff.append("&time=" + URLUTF8Encoder.encode(XMLUtils.timeDateValue(timeStamp.getTime())));
		if(location != null)
			bff.append("&l=" + location.toString());
		if(userTags != null)
			bff.append("&tag=" + URLUTF8Encoder.encode(TagsHelper.getSavableTagsString(userTags)));
		if(automaticTags != null)
			bff.append("&autotag=" + URLUTF8Encoder.encode(TagsHelper.getSavableTagsString(automaticTags)));		
		return bff.toString();
	}
    
	/**
	 * JSON representation ["localtime(YYYY-MM-DDThh:mm:ss)",db,"location","tags space separated","autotags space separated"]
	 */
	public String toJSON()
	{
		StringBuffer bff = new StringBuffer("[");
		bff.append("\"" + XMLUtils.timeDateValue( timeStamp.getTime()) + "\",");
		bff.append((int)LeqDBA.doubleValue() + ",");
		bff.append("\"" + (location != null ? location.toString() : "") + "\",");
		bff.append("\"" + (userTags != null ? TagsHelper.getSavableTagsString(userTags) : "") + "\",");
		bff.append("\"" + (automaticTags != null ? TagsHelper.getSavableTagsString(automaticTags) : "") + "\"");
		bff.append("]");
		return bff.toString();
	}
	
	/**
	 * XML representation
	 */
	public String toXML()
	{
		CustomStringBuffer bff = new CustomStringBuffer();
		bff.append("<measurement timeStamp=\"" + XMLUtils.timeDateValue(timeStamp.getTime()) + "\"");
		bff.append(" loudness=\"" + (int)LeqDBA.doubleValue() + "\"");
		if(location != null)
			bff.append(" location=\"" + location.toString() + "\"");
		if(userTags != null)
			bff.append(" tags=\"" + TagsHelper.getSavableTagsString(userTags) + "\"");
		if(automaticTags != null)
			bff.append(" autotags=\"" + TagsHelper.getSavableTagsString(automaticTags) + "\"");
		bff.append("/>");
		return bff.toString();
	}
	
}
