/**
 * --------------------------------------------------------------------------------
 *  NoiseTube Mobile client (Java implementation; Android version)
 *  
 *  Copyright (C) 2008-2010 SONY Computer Science Laboratory Paris
 *  Portions contributed by Vrije Universiteit Brussel (BrusSense team), 2008-2011
 *  Android port by Vrije Universiteit Brussel (BrusSense team), 2010-2011
 * --------------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU Lesser General Public License, version 2.1, as published
 *  by the Free Software Foundation.
 *  
 *  This library is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License along
 *  with this library; if not, write to:
 *    Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor,
 *    Boston, MA  02110-1301, USA.
 *  
 *  Full GNU LGPL v2.1 text: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
 *  NoiseTube project source code repository: http://code.google.com/p/noisetube
 * --------------------------------------------------------------------------------
 *  More information:
 *   - NoiseTube project website: http://www.noisetube.net
 *   - Sony Computer Science Laboratory Paris: http://csl.sony.fr
 *   - VUB BrusSense team: http://www.brussense.be
 * --------------------------------------------------------------------------------
 */

package net.noisetube.config.android;

import java.util.Iterator;

import android.content.SharedPreferences;

import net.noisetube.audio.calibration.Calibration;
import net.noisetube.config.NTAccount;
import net.noisetube.config.Preferences;
import net.noisetube.core.NTClient;
import net.noisetube.core.android.AndroidNTClient;
import net.noisetube.core.android.AndroidNTService;

/**
 * Android specific Preferences class
 * 
 * @author mstevens, sbarthol
 * 
 */
public class AndroidPreferences extends Preferences
{

	private static final String USER_CALIBRATION = "calibration";
	private static final String NT_ACCOUNT = "serializedAccount";
	//private static final String USE_DOSE_METER = "useDoseMeter";
	private static final String TAGS = "tags";
	private static final String USE_GPS = "useGPS";
	private static final String FORCE_GPS = "forceGPS";
	private static final String USE_INTERPOLATION = "useInterpolation";
	private static final String PAUSE_BACKGROUND = "pauseBackground";
	private static final String USE_BATCH_HTTP = "useBatchHTTP";
	private static final String SAVE_TO_FILE_WHEN_HTTP = "saveInHTTP";
	private static final String SAVING_MODE = "savingMode";
	private static final String PREFER_MEM = "preferMem";
	private static final String TAG_SEPARATOR = "";
	private static final String PREFS_NAME = "noisetube" + AndroidNTClient.CLIENT_VERSION;

	private SharedPreferences settings = null;

	@Override
	protected void setDefaults()
	{
		super.setDefaults(); //!!!

		alwaysUseBatchModeForHTTP = true;
		forceGPS = true; //don't suspend to save power
	}

	@Override
	public void loadFromStorage()
	{	
		try
		{
			//Restore preferences
			settings = AndroidNTService.getInstance().getSharedPreferences(PREFS_NAME, 0);

			//Account
			String serializedAccount = settings.getString(NT_ACCOUNT, null);
			if(serializedAccount != null)
				account = NTAccount.deserialise(serializedAccount);

			//IO
			setPreferMemoryCard(settings.getBoolean(PREFER_MEM, preferMemoryCard));
			setSavingMode(settings.getInt(SAVING_MODE, savingMode));
			setAlsoSaveToFileWhenInHTTPMode(settings.getBoolean(SAVE_TO_FILE_WHEN_HTTP, alsoSaveToFileWhenInHTTPMode));
			setAlwaysUseBatchModeForHTTP(settings.getBoolean(USE_BATCH_HTTP, alwaysUseBatchModeForHTTP));

			//GPS
			setUseGPS(settings.getBoolean(USE_GPS, useGPS));
			setForceGPS(settings.getBoolean(FORCE_GPS, forceGPS));
			setUseCoordinateInterpolation(settings.getBoolean(USE_INTERPOLATION, useCoordinateInterpolation));

			//UI
			setPauseWhenInBackground(settings.getBoolean(PAUSE_BACKGROUND, pauseWhenInBackground));

			//TAGS
			for(String tag : settings.getString(TAGS, "").split(TAG_SEPARATOR))
				if(!tag.equals(""))
					previousTags.add(tag);

			//AUDIO
			//setUseDoseMeter(settings.getBoolean(USE_DOSE_METER, false));
			String calibrationString = settings.getString(USER_CALIBRATION, null);
			if(calibrationString != null)
			{
				Calibration parsedCal = NTClient.getInstance().getCalibrationParser().parseCalibration(calibrationString, Calibration.SOURCE_USER_PREFERECES);
				if(parsedCal != null)
					calibration = parsedCal;
			}
		}
		catch (Exception e)
		{
			log.error(e, "AndroidPreferences.java -- loadFromStorage");
		}
	}

	@Override
	public void saveToStorage()
	{
		try
		{
			if(settings == null)
				settings = AndroidNTService.getInstance().getSharedPreferences(PREFS_NAME, 0);
			SharedPreferences.Editor editor = settings.edit();

			//Account
			if(account != null)
				editor.putString(account.serialise(), NT_ACCOUNT);
			
			//IO
			editor.putBoolean(PREFER_MEM, isPreferMemoryCard());
			editor.putInt(SAVING_MODE, getSavingMode());
			editor.putBoolean(SAVE_TO_FILE_WHEN_HTTP, isAlsoSaveToFileWhenInHTTPMode());
			editor.putBoolean(USE_BATCH_HTTP, isAlwaysUseBatchModeForHTTP());	

			//GPS
			editor.putBoolean(USE_GPS, isUseGPS());
			editor.putBoolean(FORCE_GPS, isForceGPS());
			editor.putBoolean(USE_INTERPOLATION, isUseCoordinateInterpolation());			

			//UI
			editor.putBoolean(PAUSE_BACKGROUND, isPauseWhenInBackground());

			//TAGS
			Iterator<String> it = previousTags.iterator();
			StringBuffer bff = new StringBuffer();
			while(it.hasNext())
			{
				bff.append(it.next());
				if(it.hasNext())
					bff.append(TAG_SEPARATOR);
			}
			editor.putString(TAGS, bff.toString());

			//AUDIO
			//editor.putBoolean(USE_DOSE_METER, isUseDoseMeter());
			if(calibration != null && calibration.isManuallyChanged())
				editor.putString(USER_CALIBRATION,calibration.toXML());

			//Commit the edits!
			editor.commit();
			System.out.println("prefs saved!");
		}
		catch (Exception e)
		{
			log.error(e, "saveToStorage()");
		}
	}
	
	/*
	 * Old file-based storage methods
	 */
	/*protected void readSettingsFrom(FileReader fr) throws IOException
	{
		char[] buf = new char[500];
		fr.read(buf);

		String buffer = new String(buf);
		buffer = buffer.trim();
		String bufferPart;
		int nextNewline;

		//Account
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		if(bufferPart != null && !bufferPart.equals("") && !bufferPart.equals(NOT_LOGGED_IN))
			account = NTAccount.deserialise(bufferPart);

		//IO (1)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		boolean preferMem = Boolean.parseBoolean(bufferPart);
		setPreferMemoryCard(preferMem);

		//IO (2)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setSavingMode(Integer.parseInt(bufferPart));

		//IO (3)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setAlsoSaveToFileWhenInHTTPMode(Boolean.parseBoolean(bufferPart));

		//IO (4)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setAlwaysUseBatchModeForHTTP(Boolean.parseBoolean(bufferPart));

		//GPS (1)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setUseGPS(Boolean.parseBoolean(bufferPart));

		//GPS (2)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setForceGPS(Boolean.parseBoolean(bufferPart));

		//GPS (3)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setUseCoordinateInterpolation(Boolean.parseBoolean(bufferPart));

		//UI
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setPauseWhenInBackground(Boolean.parseBoolean(bufferPart));

		//TAGS
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		addTags(bufferPart.split(TAG_SEPARATOR));

		//AUDIO (1)
		nextNewline = buffer.indexOf("\n");
		bufferPart = buffer.substring(0, nextNewline);
		buffer = buffer.substring(nextNewline+1);
		setUseDoseMeter(Boolean.parseBoolean(bufferPart));

		//AUDIO (2)
		bufferPart = buffer;
		if(bufferPart != null && !bufferPart.equals("") && !bufferPart.equals(UNCHANGED_CALIBRATION))
		{	
			Calibration parsedCal = NTClient.getInstance().getCalibrationParser().parseCalibration(bufferPart, Calibration.SOURCE_USER_PREFERECES);
			if(parsedCal != null)
				calibration = parsedCal;
		}

		fr.close();
	}*/

	/*protected void writeSettingsTo(FileWriter fw) throws IOException
	{
		//Account
		if(account != null)
			fw.write(account.serialise());
		else
			fw.write(NOT_LOGGED_IN);
		fw.write("\n");

		//IO
		fw.write(new Boolean(isPreferMemoryCard()).toString()); fw.write("\n");
		fw.write(""+getSavingMode()); fw.write("\n");
		fw.write(new Boolean(isAlsoSaveToFileWhenInHTTPMode()).toString()); fw.write("\n");
		fw.write(new Boolean(isAlwaysUseBatchModeForHTTP()).toString()); fw.write("\n");

		//GPS
		fw.write(new Boolean(isUseGPS()).toString()); fw.write("\n");
		fw.write(new Boolean(isForceGPS()).toString()); fw.write("\n");
		fw.write(new Boolean(isUseCoordinateInterpolation()).toString()); fw.write("\n");

		//UI
		fw.write(new Boolean(isPauseWhenInBackground()).toString()); fw.write("\n");

		//TAGS
		Iterator<String> it = previousTags.iterator();
		while(it.hasNext())
		{
			fw.write(it.next());
			if(it.hasNext())
				fw.write(TAG_SEPARATOR);
		}
		fw.write("\n");

		//AUDIO
		fw.write(new Boolean(isUseDoseMeter()).toString()); fw.write("\n");
		if(calibration != null && calibration.isManuallyChanged())
			fw.write(calibration.toXML());
		else
			fw.write(UNCHANGED_CALIBRATION);

		fw.flush();
		fw.close();
	}*/
	
}
