/**
 * --------------------------------------------------------------------------------
 *  NoiseTube Mobile client (Java implementation; Android version)
 *  
 *  Copyright (C) 2008-2010 SONY Computer Science Laboratory Paris
 *  Portions contributed by Vrije Universiteit Brussel (BrusSense team), 2008-2011
 *  Android port by Vrije Universiteit Brussel (BrusSense team), 2010-2011
 * --------------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU Lesser General Public License, version 2.1, as published
 *  by the Free Software Foundation.
 *  
 *  This library is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License along
 *  with this library; if not, write to:
 *    Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor,
 *    Boston, MA  02110-1301, USA.
 *  
 *  Full GNU LGPL v2.1 text: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
 *  NoiseTube project source code repository: http://code.google.com/p/noisetube
 * --------------------------------------------------------------------------------
 *  More information:
 *   - NoiseTube project website: http://www.noisetube.net
 *   - Sony Computer Science Laboratory Paris: http://csl.sony.fr
 *   - VUB BrusSense team: http://www.brussense.be
 * --------------------------------------------------------------------------------
 */

package net.noisetube.core.android;

import net.noisetube.R;
import net.noisetube.config.Preferences;
import net.noisetube.config.android.AndroidPreferences;
import net.noisetube.core.NTClient;
import net.noisetube.model.NTMeasurement;
import net.noisetube.model.Track;
import net.noisetube.ui.TrackUI;
import net.noisetube.ui.android.MainActivity;
import net.noisetube.ui.android.Toaster;
import net.noisetube.util.Logger;
import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.ContextWrapper;
import android.content.Intent;
import android.os.AsyncTask;
import android.os.IBinder;
import android.telephony.PhoneStateListener;
import android.telephony.TelephonyManager;

/**
 * @author sbarthol, mstevens
 */
public class AndroidNTService extends Service implements TrackUI
{

	static private AndroidNTService instance;
	private Track track = null;
	private boolean pausedForBackground = false;
	private boolean pausedForCall = false;
	private AndroidPreferences preferences;
	private MainActivity gui;
	private Logger log;
	private static final int NOTIFICATION_ID = 1;


	static public AndroidNTService getInstance()
	{
		return instance;
	}

	/** Called when the service is first created.
	 * @param savedInstanceState
	 */
	@Override
	public void onCreate()
	{	
		super.onCreate();

		if(instance == null)
			instance = this;

		//Get log instance:
		log = Logger.getInstance();

		//Register PhoneStateListener to pause measuring on incoming calls
		try
		{
			TelephonyManager telephonyManager = (TelephonyManager) getSystemService(TELEPHONY_SERVICE);
			telephonyManager.listen(new NTPhoneStateListener(), PhoneStateListener.LISTEN_CALL_STATE);
		}
		catch(Exception ignore) { } //this is a non-vital feature anyway

		/* This is necessary since the service must let the mainActivity know when it is finished,
		 * so the mainActivity is then able to bind itself to this service.
		 * It has to be done this way since the onCreate() function of this service is called asynchronously.
		 * Or in other words, mainActivity starts this service and is only able to bind itself to it once the service
		 * is fully created. Since this creation happens asynchronously, a solution that makes sure that
		 * mainActivity only starts its binding and further creation is by letting this service
		 * inform mainActivity itself when it is created.
		 * 
		 * Note that this forms a problem for when the service is created by other GuiActivity
		 * instances. Therefore it would be 'nicer' the onCreate function was able to
		 * deduce which GuiActivity started it and use this reference to contact the right GuiActivity.		 * 	
		 */
		MainActivity mainActivity = MainActivity.getInstance();
		if(mainActivity!=null)
			mainActivity.serviceCreated();		
	}

	public void bindGui(MainActivity gui)
	{
		this.gui = gui;
		destroyNotification();
	}

	private void createNotification()
	{
		try
		{
			String ns = Context.NOTIFICATION_SERVICE;
			NotificationManager mNotificationManager = (NotificationManager) getSystemService(ns);
			int icon = R.drawable.noisetube_statusbar_icon;
			CharSequence tickerText = (isPaused() ? "Paused" : "Measuring");
			long when = System.currentTimeMillis();
			Notification notification = new Notification(icon, tickerText, when);
			Context context = getApplicationContext();
			CharSequence contentTitle = "NoiseTube";
			CharSequence contentText = "Noisetube is " + (isPaused() ? "paused" : "measuring");

			Intent mainActivityIntent = new Intent(this, MainActivity.class);
			PendingIntent contentIntent = PendingIntent.getActivity(this, 0, mainActivityIntent, 0/*PendingIntent.FLAG_CANCEL_CURRENT*/);
			notification.setLatestEventInfo(context, contentTitle, contentText, contentIntent);
			//To denote that it is an ongoing app
			notification.flags |= Notification.FLAG_ONGOING_EVENT;
			//To make sure it can't be cleared using the "clear notifications" button
			notification.flags |= Notification.FLAG_NO_CLEAR;
			mNotificationManager.notify(NOTIFICATION_ID, notification);
		} catch(Exception ignore){ }
	}

	private void destroyNotification()
	{
		try
		{
			String ns = Context.NOTIFICATION_SERVICE;
			NotificationManager mNotificationManager = (NotificationManager) getSystemService(ns);
			mNotificationManager.cancel(NOTIFICATION_ID);
		} catch(Exception ignore){ }
	}

	public void initializeService()
	{
		/*
		if(gui != null)
			new InitializeAndroidNTClient().execute(gui.getContextWrapper());
		 */
		new InitializeAndroidNTClient().execute(this);
	}

	/**
	 * @author mstevens
	 *
	 */
	public class InitializeAndroidNTClient extends AsyncTask<ContextWrapper, Void, AndroidNTClient>
	{

		Exception error = null;

		@Override
		protected AndroidNTClient doInBackground(ContextWrapper... wrappers)
		{
			try
			{
				return AndroidNTClient.getInstance(wrappers[0]);
			}
			catch(Exception e)
			{
				error = e;
				return null;
			}
		}

		@Override
		protected void onPostExecute(AndroidNTClient ntClient)
		{
			//TODO display this failure to the user
			/*
			if(ntClient == null)
			{
				removeSplashScreen();
				showErrorDialog((error != null ? error : new Exception("Could not create NTClient object")));
			}
			else
			 */
			if(ntClient != null)
				finishCreation(ntClient); //will remove splash screen
		}

	}

	private void finishCreation(NTClient ntClient)
	{
		preferences = (AndroidPreferences) ntClient.getPreferences();	

		//enable file mode in the logger:
		if(preferences != null)
		{
			log.enableFileMode();
			log.disableLogBuffer(); //we don't need to log buffer now that all messages are directly saved to file
		}

		if(gui!=null)
			gui.serviceInitialized();
	}
	
	public void startMeasuring()
	{
		track = new Track();
		track.setUI(this);
		track.start();
	}

	/**
	 * Assumes there is a track and it is running (not checked here but at all calls)
	 * 
	 * @param restartAfterwards
	 */
	public void stopMeasuring(boolean restartAfterwards)
	{
		track.stop();
	}
	
	public void start_stop()
	{
		if(isMeasuring())
			stopMeasuring(false);
		else
			startMeasuring();
	}

	public void pauseMeasuringForBackground(/*boolean restart*/)
	{
		if(preferences != null && preferences.isPauseWhenInBackground())
		{
			pauseMeasuring(false);
			pausedForBackground = true;
		}
	}

	public void pauseMeasuring(boolean preStop)
	{
		if(track != null)
			track.pause(preStop);
	}

	public void resumeMeasuringFromBackground()
	{
		if(pausedForBackground && preferences.isPauseWhenInBackground())
		{
			resumeMeasuring();
			pausedForBackground = false;
		}
	}

	public void resumeMeasuring()
	{
		if(track != null)
			track.resume();
	}

	public void pause_resume()
	{
		if(track != null)
		{
			if(track.isPaused())
				track.resume();
			else
				track.pause();
		}
	}

	public void destroy()
	{
		instance = null;
		destroyNotification();
		if(isMeasuring())
		{
			track.stop(true); //!!! block until saving is done or waiting time is up (silent, does not update gui)
			Toaster.displayShortToast("Measuring stopped");
		}
		log.info("Exiting NoiseTube Mobile\n\n");
		log.disableFileMode();
		Toaster.displayToast("Goodbye");
		//Make sure the client is properly disposed of (static variables are set to null etc).
		NTClient.dispose();
		stopSelf();
	}

	/**
	 * @return the current track
	 */
	public Track getTrack()
	{
		return track;
	}

	public AndroidPreferences getPreferences()
	{
		return preferences;
	}

	public boolean isMeasuring()
	{
		return track != null && track.isRunning();
	}
	
	public boolean isPaused()
	{
		return track != null && track.isPaused();
	}

	public class NTPhoneStateListener extends PhoneStateListener
	{

		@Override
		public void onCallStateChanged(int state, String incomingNumber)
		{
			if(track == null)
				return;
			try
			{
				switch(state)
				{
				case TelephonyManager.CALL_STATE_RINGING : track.pause(); pausedForCall = true; break;
				case TelephonyManager.CALL_STATE_OFFHOOK : track.pause(); pausedForCall = true; break;
				case TelephonyManager.CALL_STATE_IDLE :
					if(pausedForCall)
					{	
						track.resume();
						pausedForCall = false;
					}
					break;
				}
			}
			catch(Exception ignore) { }
		}

	}

	public boolean loginRequired() //TODO move this to NTCLient?
	{
		return (preferences!=null && preferences.getSavingMode() == Preferences.SAVE_HTTP && !preferences.isAuthenticated());
	}

	@Override
	public IBinder onBind(Intent arg0) 
	{
		// TODO Auto-generated method stub
		return null;
	}

	public void newMeasurement(Track track, NTMeasurement newMeasurement, NTMeasurement savedMeasurement) 
	{
		if(gui!=null)
			gui.newMeasurement(track, newMeasurement, savedMeasurement);
	}

	public void measuringStarted(Track track) 
	{
		if(gui!=null)
			gui.measuringStarted(track);
	}

	public void measuringPaused(Track track) 
	{
		if(gui!=null)
			gui.measuringPaused(track);
	}

	public void measuringResumed(Track track)
	{
		if(gui!=null)
			gui.measuringResumed(track);
	}

	public void measuringStopped(Track track) 
	{
		if(gui!=null)
			gui.measuringStopped(track);
	}

	public void decouple()
	{
		gui = null;
		createNotification();
	}

}
