/**
 * --------------------------------------------------------------------------------
 *  NoiseTube Mobile client (Java implementation; Android version)
 *  
 *  Copyright (C) 2008-2010 SONY Computer Science Laboratory Paris
 *  Portions contributed by Vrije Universiteit Brussel (BrusSense team), 2008-2011
 *  Android port by Vrije Universiteit Brussel (BrusSense team), 2010-2011
 * --------------------------------------------------------------------------------
 *  This library is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU Lesser General Public License, version 2.1, as published
 *  by the Free Software Foundation.
 *  
 *  This library is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 *  details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License along
 *  with this library; if not, write to:
 *    Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor,
 *    Boston, MA  02110-1301, USA.
 *  
 *  Full GNU LGPL v2.1 text: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
 *  NoiseTube project source code repository: http://code.google.com/p/noisetube
 * --------------------------------------------------------------------------------
 *  More information:
 *   - NoiseTube project website: http://www.noisetube.net
 *   - Sony Computer Science Laboratory Paris: http://csl.sony.fr
 *   - VUB BrusSense team: http://www.brussense.be
 * --------------------------------------------------------------------------------
 */

package net.noisetube.ui.android;

import java.util.Collections;
import java.util.List;
import java.util.Vector;

import net.noisetube.R;
import net.noisetube.config.android.AndroidPreferences;
import net.noisetube.core.android.AndroidNTService;
import net.noisetube.model.TaggedInterval;
import net.noisetube.model.Track;
import net.noisetube.tagging.TagsHelper;
import net.noisetube.ui.android.views.MultiSpinner;
import net.noisetube.ui.android.views.MultiSpinner.MultiSpinnerItem;
import net.noisetube.ui.android.views.MultiSpinner.MultiSpinnerListener;
import net.noisetube.util.Logger;
import android.app.Dialog;
import android.content.Context;
import android.os.Bundle;
import android.text.Editable;
import android.text.TextWatcher;
import android.view.View;
import android.widget.Button;
import android.widget.EditText;

/**
 * @author sbarthol, mstevens
 *
 */
public class TagDialog extends Dialog
{
	
	private EditText txtTags;
	private MultiSpinner tagSpinner;
	private Button btnTag;

	private int numberFirstM = 0;
	private int numberLastM = 0;

	private Track track;
	
	public TagDialog(Context context, Track track, int numberFirstM, int numberLastM)
	{
		super(context);
		this.track = track;
		this.numberFirstM = numberFirstM;
		this.numberLastM = numberLastM;
	}

	@Override
	public void onCreate(Bundle savedInstanceState)
	{
		super.onCreate(savedInstanceState);
		
		final AndroidPreferences preferences = AndroidNTService.getInstance().getPreferences();

		setContentView(R.layout.tagdialog);
		setTitle(R.string.tag_dialog_title);

		/* get child views */
		txtTags = (EditText) findViewById(R.id.txtTags);
		btnTag = (Button) findViewById(R.id.tagsButton);
		tagSpinner = (MultiSpinner) findViewById(R.id.tagsSpinner);
		
		/* configure them */
		//Tag text box
		txtTags.setFocusable(true);
		txtTags.setFocusableInTouchMode(true);
		txtTags.addTextChangedListener(new TextWatcher()
		{
			
			public void onTextChanged(CharSequence s, int start, int before, int count)
			{
				if(s.length() > 0 || tagSpinner.isSomethingSelected())
					btnTag.setText(getContext().getString(R.string.tag_dialog_btn_apply));
				else
					btnTag.setText(getContext().getString(R.string.tag_dialog_btn_cancel));
			}
			
			public void beforeTextChanged(CharSequence s, int start, int count, int after) {}
			
			public void afterTextChanged(Editable s) {}
		});
		
		//Spinner with previous tags
		if(preferences.getTags().isEmpty())
		{
			tagSpinner.setDefaultText(getContext().getString(R.string.tag_dialog_spinner_text_empty));
			tagSpinner.setEnabled(false);
		}
		else
		{
			tagSpinner.setEnabled(true);
			tagSpinner.setDefaultText(getContext().getString(R.string.tag_dialog_spinner_text));
			tagSpinner.setSeparator(" ");
			Collections.sort(preferences.getTags()); //put tags in alphabetical order
			for(String oldTag : (Vector<String>) preferences.getTags())
				tagSpinner.addItem(oldTag, false);
		}
		tagSpinner.setListener(new MultiSpinnerListener()
		{
			
			public void onSelectionChanged(MultiSpinner spinner, List<MultiSpinnerItem> items)
			{
				if(spinner.isSomethingSelected() || txtTags.getText().length() > 0)
					btnTag.setText(getContext().getString(R.string.tag_dialog_btn_apply));
				else
					btnTag.setText(getContext().getString(R.string.tag_dialog_btn_cancel));
			}
		});
		
		//Apply/cancel button
		btnTag.setText(getContext().getString(R.string.tag_dialog_btn_cancel));
		btnTag.setOnClickListener(new View.OnClickListener()
		{
			//adds the tag that can currently be found to the tag-variable, to the stored measurements.
			public void onClick(View btn)
			{
				try
				{
					Vector<String> tags = new Vector<String>();
					//typed tags:
					for(String tag : TagsHelper.splitTypedTags(txtTags.getText().toString()))
						if(!tag.equals(""))
							tags.add(tag);
					txtTags.setText(""); //clear field
					
					//selected tags:
					for(MultiSpinnerItem i : tagSpinner.getItems())
						if(i.isSelected())
							tags.add(i.toString());
					if(track!=null && !tags.isEmpty())
					{
						//insert tagged interval:
						track.addTaggedInterval(new TaggedInterval(numberFirstM, numberLastM, tags));
						
						//save tags for later:
						for(String tag : tags)
						{
							if(!preferences.getTags().contains(tag))
								preferences.addTag(tag);
						}
						preferences.saveToStorage();
						
						//inform user
						String tagString = "";
						for(String tag : tags)
							tagString += (tagString.length() > 0 ? " " : "") + tag;
						//+1 since numberFirstM and numberLastM start counting from 0, while most users feel that counting starting with 1 is more natural
						Toaster.displayToast("Measurements number " + (numberFirstM+1) + " up until number " + (numberLastM+1) + " have been tagged with \"" + tagString + "\"");
					}
					
					//Always resume measuring after tagging (also if it was paused before the selection of the interval):
					AndroidNTService.getInstance().resumeMeasuring();
					cancel(); //will trigger OnCancel event
				}
				catch(Exception e)
				{
					Logger.getInstance().error(e, "upon apply in TagDialog");
					cancel();
					return;
				}
			}
		});
	}
	
	public void onBackPressed()
	{
		//Always resume measuring after tagging (also if it was paused before the selection of the interval):
		AndroidNTService.getInstance().resumeMeasuring();
		cancel();
	}

}
